import yacron.job
import yacron.config
import asyncio
import pytest
import aiosmtplib
import raven
from unittest.mock import Mock, patch


@pytest.mark.parametrize("save_limit, input_lines, output, expected_failure", [
    (10,
     b"line1\nline2\nline3\nline4\n",
     'line1\nline2\nline3\nline4\n',
     True),
    (1,
     b"line1\nline2\nline3\nline4\n",
     '   [.... 3 lines discarded ...]\nline4\n',
     True),
    (2,
     b"line1\nline2\nline3\nline4\n",
     'line1\n   [.... 2 lines discarded ...]\nline4\n',
     True),
    (0, b"line1\nline2\nline3\nline4\n", '', True),
    (0, b"", '', False),
])
def test_stream_reader(save_limit, input_lines, output, expected_failure):
    loop = asyncio.get_event_loop()
    fake_stream = asyncio.StreamReader()
    reader = yacron.job.StreamReader("cronjob-1", "stderr", fake_stream,
                                     save_limit)

    job_config = yacron.config.parse_config_string('''
jobs:
  - name: test
    command: foo
    schedule: "* * * * *"
    captureStderr: true
''')[0]
    job = yacron.job.RunningJob(job_config, None)

    async def producer(fake_stream):
        fake_stream.feed_data(input_lines)
        fake_stream.feed_eof()

    job._stderr_reader = reader
    job.retcode = 0
    loop.run_until_complete(asyncio.gather(
        producer(fake_stream),
        job._read_job_streams()))
    out = job.stderr

    assert (out, job.failed) == (output, expected_failure)


A_JOB = '''
jobs:
  - name: test
    command: ls
    schedule: "* * * * *"
    onSuccess:
      report:
        mail:
          from: example@foo.com
          to: example@bar.com
          smtpHost: smtp1
          smtpPort: 1025
          subject: "Cron job '{{name}}' {% if success %}completed{% else %}failed{% endif %}"
          body: |
            {% if stdout and stderr -%}
            STDOUT:
            ---
            {{stdout}}
            ---
            STDERR:
            {{stderr}}
            {% elif stdout -%}
            {{stdout}}
            {% elif stderr -%}
            {{stderr}}
            {% else -%}
            (no output was captured)
            {% endif %}
'''


@pytest.mark.parametrize("success, stdout, stderr, subject, body", [
    (True, "out", "err",
     "Cron job 'test' completed",
     'STDOUT:\n---\nout\n---\nSTDERR:\nerr\n'),

    (False, "out", "err",
     "Cron job 'test' failed",
     'STDOUT:\n---\nout\n---\nSTDERR:\nerr\n'),

    (False, None, None,
     "Cron job 'test' failed",
     "(no output was captured)\n"),

    (False, None, "err",
     "Cron job 'test' failed",
     'err\n'),

    (False, "out", None,
     "Cron job 'test' failed",
     'out\n'),
])
def test_report_mail(success, stdout, stderr, subject, body):
    job_config = yacron.config.parse_config_string(A_JOB)[0]
    print(job_config.onSuccess['report'])
    job = Mock(config=job_config, stdout=stdout, stderr=stderr,
               template_vars={
                   'name': job_config.name,
                   'success': success,
                   'stdout': stdout,
                   'stderr': stderr,
               })

    mail = yacron.job.MailReporter()
    loop = asyncio.get_event_loop()

    connect_calls = []
    messages_sent = []

    async def connect(self):
        connect_calls.append(self)

    async def send_message(self, message):
        messages_sent.append(message)

    real_init = aiosmtplib.SMTP.__init__
    smtp_init_args = None

    def init(self, *args, **kwargs):
        nonlocal smtp_init_args
        smtp_init_args = args, kwargs
        real_init(self, *args, **kwargs)

    with patch("aiosmtplib.SMTP.__init__", init), \
            patch("aiosmtplib.SMTP.connect", connect), \
            patch("aiosmtplib.SMTP.send_message", send_message):
        loop.run_until_complete(mail.report(success,
                                            job,
                                            job_config.onSuccess['report']))

    assert smtp_init_args == ((), {'hostname': 'smtp1', 'port': 1025})
    assert len(connect_calls) == 1
    assert len(messages_sent) == 1
    message = messages_sent[0]
    assert message['From'] == "example@foo.com"
    assert message['To'] == "example@bar.com"
    assert message['Subject'] == subject
    assert message.get_payload() == body


@pytest.mark.parametrize(
    "success, dsn_from, body, extra, expected_dsn, fingerprint", [
    (True,
     "value",
     "Cron job 'test' completed\n\nSTDOUT:\n---\nout\n---\nSTDERR:\nerr\n",
     {
        'job': 'test',
        'exit_code': 0,
        'command': 'ls',
        'shell': '/bin/sh',
        'success': True,
    }, "http://xxx:yyy@sentry/1", ["test"]),
    (False,
     "file",
     "Cron job 'test' failed\n\nSTDOUT:\n---\nout\n---\nSTDERR:\nerr\n",
     {
        'job': 'test',
        'exit_code': 0,
        'command': 'ls',
        'shell': '/bin/sh',
        'success': False,
    }, "http://xxx:yyy@sentry/2", ["test"]),
    (False,
     "envvar",
     "Cron job 'test' failed\n\nSTDOUT:\n---\nout\n---\nSTDERR:\nerr\n",
     {
        'job': 'test',
        'exit_code': 0,
        'command': 'ls',
        'shell': '/bin/sh',
        'success': False,
    }, "http://xxx:yyy@sentry/3", ["test"]),
])
def test_report_sentry(success, dsn_from, body, extra, expected_dsn,
                       fingerprint, tmpdir, monkeypatch):
    job_config = yacron.config.parse_config_string(A_JOB)[0]

    p = tmpdir.join("sentry-secret-dsn")
    p.write("http://xxx:yyy@sentry/2")

    monkeypatch.setenv("TEST_SENTRY_DSN", "http://xxx:yyy@sentry/3")

    if dsn_from == 'value':
        job_config.onSuccess['report']['sentry'] = {
            "dsn": {
                "value": "http://xxx:yyy@sentry/1",
                "fromFile": None,
                "fromEnvVar": None,
            },
            'body': (yacron.config.DEFAULT_CONFIG['onFailure']['report']
                     ['sentry']['body']),
            'fingerprint': ['{{ name }}'],
        }
    elif dsn_from == 'file':
        job_config.onSuccess['report']['sentry'] = {
            "dsn": {
                "value": None,
                "fromFile": str(p),
                "fromEnvVar": None,
            },
            'body': (yacron.config.DEFAULT_CONFIG['onFailure']['report']
                     ['sentry']['body']),
            'fingerprint': ['{{ name }}'],
        }
    elif dsn_from == 'envvar':
        job_config.onSuccess['report']['sentry'] = {
            "dsn": {
                "value": None,
                "fromFile": None,
                "fromEnvVar": "TEST_SENTRY_DSN",
            },
            'body': (yacron.config.DEFAULT_CONFIG['onFailure']['report']
                     ['sentry']['body']),
            'fingerprint': ['{{ name }}'],
        }
    else:
        raise AssertionError

    job = Mock(config=job_config, stdout="out", stderr="err", retcode=0,
               template_vars={
                   'name': job_config.name,
                   'success': success,
                   'stdout': "out",
                   'stderr': "err",
               })
    loop = asyncio.get_event_loop()

    messages_sent = []

    def captureMessage(self, body, extra, *, fingerprint):
        messages_sent.append((body, extra, fingerprint))

    real_init = raven.Client.__init__
    init_args = (), {}

    def init(self, *args, **kwargs):
        nonlocal init_args
        init_args = args, kwargs
        real_init(self, *args, **kwargs)

    sentry = yacron.job.SentryReporter()
    with patch("raven.Client.__init__", init), \
            patch("raven.Client.captureMessage", captureMessage):
        loop.run_until_complete(sentry.report(success,
                                              job,
                                              job_config.onSuccess['report']))

    args, kwargs = init_args
    assert kwargs.get('dsn') == expected_dsn

    assert len(messages_sent) == 1
    got_body, got_extra, got_fingerprint = messages_sent[0]
    assert got_body == body
    assert got_extra == extra
    assert got_fingerprint == fingerprint


@pytest.mark.parametrize("shell, command, expected_type, expected_args", [
    ('', "Civ 6", 'shell', ('Civ 6',)),
    ('', ["echo", "hello"], 'exec', ('echo', 'hello')),
    ('bash', 'echo "hello"', 'exec', ('bash', '-c', 'echo "hello"',)),
])
def test_job_run(monkeypatch, shell, command, expected_type, expected_args):

    shell_commands = []
    exec_commands = []

    async def create_subprocess_common(*args, **kwargs):
        stdout = asyncio.StreamReader()
        stderr = asyncio.StreamReader()
        stdout.feed_data(b"out\n")
        stdout.feed_eof()
        stderr.feed_data(b"err\n")
        stderr.feed_eof()
        proc = Mock(stdout=stdout, stderr=stderr)

        async def wait():
            return

        proc.wait = wait
        return proc

    async def create_subprocess_shell(*args, **kwargs):
        shell_commands.append((args, kwargs))
        return await create_subprocess_common(*args, **kwargs)

    async def create_subprocess_exec(*args, **kwargs):
        exec_commands.append((args, kwargs))
        return await create_subprocess_common(*args, **kwargs)

    monkeypatch.setattr("asyncio.create_subprocess_exec",
                        create_subprocess_exec)
    monkeypatch.setattr("asyncio.create_subprocess_shell",
                        create_subprocess_shell)

    if isinstance(command, list):
        command_snippet = '\n'.join(
            ["    command:"] + ['      - ' + arg for arg in command])
    else:
        command_snippet = '    command: ' + command

    job_config = yacron.config.parse_config_string('''
jobs:
  - name: test
{command}
    schedule: "* * * * *"
    shell: {shell}
    captureStderr: true
    captureStdout: true
    environment:
      - key: FOO
        value: bar
'''.format(command=command_snippet, shell=shell))[0]

    job = yacron.job.RunningJob(job_config, None)

    async def run(job):
        await job.start()
        await job.wait()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run(job))

    if shell_commands:
        run_type = 'shell'
        assert len(shell_commands) == 1
        args, kwargs = shell_commands[0]
    elif exec_commands:
        run_type = 'exec'
        assert len(exec_commands) == 1
        args, kwargs = exec_commands[0]
    else:
        raise AssertionError

    assert kwargs['env']['FOO'] == 'bar'
    assert run_type == expected_type
    assert args == expected_args


def test_execution_timeout():
    job_config = yacron.config.parse_config_string('''
jobs:
  - name: test
    command: |
        echo "hello"
        sleep 1
        echo "world"
    executionTimeout: 0.25
    schedule: "* * * * *"
    captureStderr: false
    captureStdout: true
''')[0]

    async def test(job):
        await job.start()
        await job.wait()
        return job.stdout

    job = yacron.job.RunningJob(job_config, None)
    loop = asyncio.get_event_loop()
    stdout = loop.run_until_complete(test(job))
    assert stdout == "hello\n"


def test_error1():
    job_config = yacron.config.parse_config_string('''
jobs:
  - name: test
    command: echo "hello"
    schedule: "* * * * *"
''')[0]
    job = yacron.job.RunningJob(job_config, None)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(job.start())
    with pytest.raises(RuntimeError):
        loop.run_until_complete(job.start())


def test_error2():
    job_config = yacron.config.parse_config_string('''
jobs:
  - name: test
    command: echo "hello"
    schedule: "* * * * *"
''')[0]
    job = yacron.job.RunningJob(job_config, None)

    loop = asyncio.get_event_loop()
    with pytest.raises(RuntimeError):
        loop.run_until_complete(job.wait())


def test_error3():
    job_config = yacron.config.parse_config_string('''
jobs:
  - name: test
    command: echo "hello"
    schedule: "* * * * *"
''')[0]
    job = yacron.job.RunningJob(job_config, None)

    loop = asyncio.get_event_loop()
    with pytest.raises(RuntimeError):
        loop.run_until_complete(job.cancel())


@pytest.mark.parametrize('command', [
    'echo "hello"',
    'exit 1',
])
def test_statsd(command):
    loop = asyncio.get_event_loop()
    received = []

    async def run():
        class UDPServerProtocol:

            def connection_made(self, transport):
                self.transport = transport

            def datagram_received(self, data, addr):
                print('Statsd UDP packet received:', data)
                message = data.decode()
                received.extend(m for m in message.split('\n') if m)

            def connection_lost(*_):
                pass

        listen = loop.create_datagram_endpoint(
            UDPServerProtocol, local_addr=('127.0.0.1', 0))
        transport, protocol = await listen

        host, port = transport.get_extra_info('sockname')
        print('Listening UDP on %s:%s' % (host, port))

        job_config = yacron.config.parse_config_string('''
jobs:
  - name: test
    command: {command}
    schedule: "* * * * *"
    statsd:
      host: 127.0.0.1
      port: {port}
      prefix: the.prefix
'''.format(port=port, command=command))[0]

        job = yacron.job.RunningJob(job_config, None)

        await job.start()
        await job.wait()
        await asyncio.sleep(0.05)
        transport.close()
        await asyncio.sleep(0.05)
        return job

    job = loop.run_until_complete(run())

    assert received
    assert len(received) == 4
    assert 'the.prefix.start' in received[0]
    assert any('the.prefix.stop' in r for r in received[1:])
    success = 0 if job.failed else 1
    assert any('the.prefix.success:%i' % success in r for r in received[1:])
    assert any('the.prefix.duration' in r for r in received[1:])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()


requirements = [
    "strictyaml>=0.7.2",
    "crontab",
    "aiohttp",
    "raven",
    "raven-aiohttp",
    "aiosmtplib",
    "jinja2",
]


setup_requirements = [
    'setuptools_scm',
    'pytest-runner',
]

test_requirements = [
    'pytest-cov',
    # https://github.com/pytest-dev/pytest-runner/issues/11#issuecomment-190355698
    'pytest',  # needs to be last
]

with open('HISTORY.rst') as history_file:
    history = history_file.read()

setup(
    name='yacron',
    version='0.1.0',
    description="A modern Cron replacement that is Docker-friendly",
    long_description=(readme + '\n\n' + history),
    author="Gustavo Carneiro",
    author_email='gustavocarneiro@gambitresearch.com',
    url='https://github.com/gjcarneiro/yacron',
    packages=find_packages(include=['yacron']),
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=True,
    keywords='yacron',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Environment :: Console',
        'Topic :: Utilities',
    ],
    setup_requires=setup_requirements,
    tests_require=test_requirements,
    use_scm_version=True,
    entry_points={
        'console_scripts': [
            'yacron = yacron.__main__:main',
        ],
    },
    python_requires='>=3.5',
)



import asyncio
import asyncio.subprocess
import logging
import os
import sys
import time
from email.mime.text import MIMEText
from socket import gethostname
from typing import Any, Awaitable, Dict, List, Optional, Union, Tuple  # noqa

from raven import Client
from raven_aiohttp import AioHttpTransport

import aiosmtplib
import jinja2

from yacron.config import JobConfig
from yacron.statsd import StatsdJobMetricWriter

logger = logging.getLogger('yacron')


if 'HOSTNAME' not in os.environ:
    os.environ['HOSTNAME'] = gethostname()


def create_task(coro: Awaitable) -> asyncio.Task:
    return asyncio.get_event_loop().create_task(coro)


class StreamReader:

    def __init__(self,
                 job_name: str,
                 stream_name: str,
                 stream: asyncio.StreamReader,
                 save_limit: int) -> None:
        self.save_top = []  # type: List[str]
        self.save_bottom = []  # type: List[str]
        self.job_name = job_name
        self.save_limit = save_limit
        self.stream_name = stream_name
        self._reader = create_task(self._read(stream))
        self.discarded_lines = 0

    async def _read(self, stream):
        prefix = "[{} {}] ".format(self.job_name, self.stream_name)
        limit_top = self.save_limit // 2
        limit_bottom = self.save_limit - limit_top
        while True:
            line = (await stream.readline()).decode("utf-8")
            if not line:
                return
            sys.stdout.write(prefix + line)
            sys.stdout.flush()
            if self.save_limit > 0:
                if len(self.save_top) < limit_top:
                    self.save_top.append(line)
                else:
                    if len(self.save_bottom) == limit_bottom:
                        del self.save_bottom[0]
                        self.discarded_lines += 1
                    self.save_bottom.append(line)
            else:
                self.discarded_lines += 1

    async def join(self) -> Tuple[str, int]:
        await self._reader
        if self.save_bottom:
            middle = (["   [.... {} lines discarded ...]\n"
                       .format(self.discarded_lines)]
                      if self.discarded_lines else [])
            output = ''.join(self.save_top + middle + self.save_bottom)
        else:
            output = ''.join(self.save_top)
        return output, self.discarded_lines


class Reporter:

    async def report(self, success: bool, job: 'RunningJob',
                     config: Dict[str, Any]) -> None:
        raise NotImplementedError  # pragma: no cover


class SentryReporter(Reporter):

    async def report(self, success: bool, job: 'RunningJob',
                     config: Dict[str, Any]) -> None:
        config = config['sentry']
        if config['dsn']['value']:
            dsn = config['dsn']['value']
        elif config['dsn']['fromFile']:
            with open(config['dsn']['fromFile'], "rt") as dsn_file:
                dsn = dsn_file.read().strip()
        elif config['dsn']['fromEnvVar']:
            dsn = os.environ[config['dsn']['fromEnvVar']]
        else:
            return  # sentry disabled: early return

        template = jinja2.Template(config['body'])
        body = template.render(job.template_vars)

        fingerprint = []
        for line in config['fingerprint']:
            fingerprint.append(jinja2.Template(line).render(job.template_vars))

        extra = {
            'job': job.config.name,
            'exit_code': job.retcode,
            'command': job.config.command,
            'shell': job.config.shell,
            'success': success,
        }
        logger.debug("sentry body: %r; fingerprint: %r", body, fingerprint)
        client = Client(transport=AioHttpTransport,
                        dsn=dsn,
                        string_max_length=4096)
        client.captureMessage(
            body,
            extra=extra,
            fingerprint=fingerprint,
        )


class MailReporter(Reporter):

    async def report(self, success: bool, job: 'RunningJob',
                     config: Dict[str, Any]) -> None:
        mail = config['mail']
        if not (mail['to'] and mail['from']):
            return  # email reporting disabled
        smtp_host = mail['smtpHost']
        smtp_port = mail['smtpPort']
        tmpl_vars = job.template_vars
        body_tmpl = jinja2.Template(mail['body'])
        body = body_tmpl.render(tmpl_vars)
        if success and not body.strip():
            logger.debug("body is empty, not sending email")
            return
        subject_tmpl = jinja2.Template(mail['subject'])
        subject = subject_tmpl.render(tmpl_vars)

        logger.debug("smtp: host=%r, port=%r", smtp_host, smtp_port)
        message = MIMEText(body)
        message['From'] = mail['from']
        message['To'] = mail['to']
        message['Subject'] = subject
        smtp = aiosmtplib.SMTP(hostname=smtp_host, port=smtp_port)
        await smtp.connect()
        await smtp.send_message(message)


class JobRetryState:

    def __init__(self, initial_delay: float,
                 multiplier: float,
                 max_delay: float) -> None:
        self.multiplier = multiplier
        self.max_delay = max_delay
        self.delay = initial_delay
        self.count = 0  # number of times retried
        self.task = None  # type: Optional[asyncio.Task]
        self.cancelled = False

    def next_delay(self) -> float:
        delay = self.delay
        self.delay = min(delay * self.multiplier, self.max_delay)
        self.count += 1
        return delay


class RunningJob:
    REPORTERS = [
        SentryReporter(),
        MailReporter(),
    ]  # type: List[Reporter]

    def __init__(self, config: JobConfig,
                 retry_state: Optional[JobRetryState]) -> None:
        self.config = config
        self.proc = None  # type: Optional[asyncio.subprocess.Process]
        self.retcode = None  # type: Optional[int]
        self._stderr_reader = None  # type: Optional[StreamReader]
        self._stdout_reader = None  # type: Optional[StreamReader]
        self.stderr = None  # type: Optional[str]
        self.stdout = None  # type: Optional[str]
        self.stderr_discarded = 0
        self.stdout_discarded = 0
        self.execution_deadline = None  # type: Optional[float]
        self.retry_state = retry_state
        self.env = None  # type: Optional[Dict[str, str]]

        statsd_config = self.config.statsd
        if statsd_config is not None:
            self.statsd_writer = StatsdJobMetricWriter(
                host=statsd_config['host'],
                port=statsd_config['port'],
                prefix=statsd_config['prefix'],
                job=self,
            )  # type: Optional[StatsdJobMetricWriter]
        else:
            self.statsd_writer = None

    async def _on_start(self) -> None:
        if self.statsd_writer:
            await self.statsd_writer.job_started()

    async def _on_stop(self) -> None:
        if self.statsd_writer:
            await self.statsd_writer.job_stopped()

    async def start(self) -> None:
        if self.proc is not None:
            raise RuntimeError("process already running")
        kwargs = {}  # type: Dict[str, Any]
        if isinstance(self.config.command, list):
            create = asyncio.create_subprocess_exec
            cmd = self.config.command
        else:
            if self.config.shell:
                create = asyncio.create_subprocess_exec
                cmd = [self.config.shell, '-c', self.config.command]
            else:
                create = asyncio.create_subprocess_shell
                cmd = [self.config.command]
        if self.config.environment:
            env = dict(os.environ)
            for envvar in self.config.environment:
                env[envvar['key']] = envvar['value']
                self.env = env
            kwargs['env'] = env
        logger.debug("%s: will execute argv %r", self.config.name, cmd)
        if self.config.captureStderr:
            kwargs['stderr'] = asyncio.subprocess.PIPE
        if self.config.captureStdout:
            kwargs['stdout'] = asyncio.subprocess.PIPE
        if self.config.executionTimeout:
            self.execution_deadline = (time.perf_counter() +
                                       self.config.executionTimeout)

        self.proc = await create(*cmd, **kwargs)

        await self._on_start()

        if self.config.captureStderr:
            assert self.proc.stderr is not None
            self._stderr_reader = \
                StreamReader(self.config.name, 'stderr', self.proc.stderr,
                             self.config.saveLimit)
        if self.config.captureStdout:
            assert self.proc.stdout is not None
            self._stdout_reader = \
                StreamReader(self.config.name, 'stdout', self.proc.stdout,
                             self.config.saveLimit)

    async def wait(self) -> None:
        if self.proc is None:
            raise RuntimeError("process is not running")
        if self.execution_deadline is None:
            self.retcode = await self.proc.wait()
            await self._on_stop()
        else:
            timeout = self.execution_deadline - time.perf_counter()
            try:
                if timeout > 0:
                    self.retcode = await asyncio.wait_for(
                        self.proc.wait(),
                        timeout,
                    )
                    await self._on_stop()
                else:
                    raise asyncio.TimeoutError
            except asyncio.TimeoutError:
                logger.info("Job %s exceeded its executionTimeout of "
                            "%.1f seconds, cancelling it...",
                            self.config.name, self.config.executionTimeout)
                self.retcode = -100
                await self.cancel()
        await self._read_job_streams()

    async def _read_job_streams(self):
        if self._stderr_reader:
            self.stderr, self.stderr_discarded = \
                await self._stderr_reader.join()
        if self._stdout_reader:
            self.stdout, self.stdout_discarded = \
                await self._stdout_reader.join()

    @property
    def failed(self) -> bool:
        if self.config.failsWhen['always']:
            return True
        if self.config.failsWhen['nonzeroReturn'] and self.retcode != 0:
            return True
        if (self.config.failsWhen['producesStdout'] and
                (self.stdout or self.stdout_discarded)):
            return True
        if (self.config.failsWhen['producesStderr'] and
                (self.stderr or self.stderr_discarded)):
            return True
        return False

    async def cancel(self) -> None:
        if self.proc is None:
            raise RuntimeError("process is not running")
        self.proc.terminate()
        try:
            await asyncio.wait_for(self.proc.wait(), self.config.killTimeout)
        except asyncio.TimeoutError:
            logger.warning("Job %s did not gracefully terminate after "
                           "%.1f seconds, killing it...",
                           self.config.name, self.config.killTimeout)
            self.proc.kill()
        await self._on_stop()

    async def report_failure(self):
        logger.info("Cron job %s: reporting failure", self.config.name)
        await self._report_common(self.config.onFailure['report'], False)

    async def report_permanent_failure(self):
        logger.info("Cron job %s: reporting permanent failure",
                    self.config.name)
        await self._report_common(self.config.onPermanentFailure['report'],
                                  False)

    async def report_success(self):
        logger.info("Cron job %s: reporting success", self.config.name)
        await self._report_common(self.config.onSuccess['report'], True)

    async def _report_common(self, report_config: dict,
                             success: bool) -> None:
        results = await asyncio.gather(
            *[reporter.report(success, self, report_config)
              for reporter in self.REPORTERS],
            return_exceptions=True
        )
        for result in results:
            if isinstance(result, Exception):
                logger.error("Problem reporting job %s failure: %s",
                             self.config.name, result)

    @property
    def template_vars(self) -> dict:
        return {
            'name': self.config.name,
            'success': not self.failed,
            'stdout': self.stdout,
            'stderr': self.stderr,
            'exit_code': self.retcode,
            'command': self.config.command,
            'shell': self.config.shell,
            'environment': self.env,
        }
